//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// ファンクションコードと Read/Write 方向の情報付きアドレス。
//
// |<--------------- attrerty ----------------->|<-- address -->|
//  63            39  38  37  36  35  34  33  32  31           0
// +---- ... ---+---+---+---+---+---+---+---+---+----- ...  ----+
// |          0 | 0 |R/W| 0 | 0 |S/U| I | D |S/U|     Address   |
// +---- ... ---+---+---+---+---+---+---+---+---+----- ...  ----+
//
// 32ビット目と35ビット目はどちらも S/U で同じ値を入れる。m68k の FC2 に相当。
// 33ビット目は D (データ空間)。m68k の FC0 に相当。
// 34ビット目は I (命令/プログラム空間)。m68k の FC1 に相当。
// これにより、
// o 下位 33 ビットで S/U と32ビットアドレスが表現できる。
// o 下位 35 ビットで FC(順不同) と32ビットアドレスが表現できる。
// o 上位ワードを >>1 すると下位3ビットが FC2-0 (同順) になる。
//
// 38ビット目は R/W (Read なら 1)。これは SSW で使う。

#pragma once

// fcaddr の上位ワード部分。
class fcattr
{
 public:
	static const uint32 D = 0x01 << 1;	// データ空間
	static const uint32 I = 0x02 << 1;	// 命令/プログラム空間
	static const uint32 S = 0x09;		// スーパーバイザ
	static const uint32 U = 0x00;		// ユーザ
	static const uint32 R = 0x40;		// 読み込み
	static const uint32 W = 0x00;		// 書き込み

	static const uint32 Fetch = I | R;
	static const uint32 Read  = D | R;
	static const uint32 Write = D | W;

 public:
	// コンストラクタ
	fcattr() {
	}
	fcattr(uint32 attr_) {
		attr = attr_;
	}

	// 代入演算子
	fcattr& operator=(const fcattr& rhs) {
		attr = rhs.attr;
		return *this;
	}

	// キャスト演算子
	explicit operator uint32() const { return attr; }

	// S/U からの変換
	static constexpr uint32 SU(bool s) {
		return (s ? S : U);
	}
	// FC からの変換
	static constexpr uint32 FC(uint32 fc) {
		return (fc << 1) | (fc >> 2);
	}

	// FC のみを取得
	uint32 GetFC() const {
		return (attr >> 1) & 7;
	}

 private:
	uint32 attr;
};

// fcattr の OR 演算子
static __unused fcattr operator|(const fcattr& lhs, const fcattr& rhs) {
	return (fcattr)((uint32)lhs | (uint32)rhs);
}


// 属性情報付きアドレス。
class fcaddr
{
 public:
	// コンストラクタ
	fcaddr() {
	}
	// コンストラクタ (属性はゼロクリア)
	fcaddr(uint32 addr_) {
		SetAddr(addr_);
	}
	// コンストラクタ (アドレスと属性)
	fcaddr(uint32 addr_, fcattr attr_) {
		Set(addr_, attr_);
	}

	// コピーコンストラクタを自前で定義すると、
	// このクラスがレジスタ渡しから参照渡しになるようなので、
	// コピーコンストラクタはデフォルトのまま使う。

	// 代入演算子
	fcaddr& operator=(const fcaddr& rhs) {
		value.q = rhs.value.q;
		return *this;
	}

	// 代入演算子 (uint32 を代入するとアドレスだけ差し替える)
	fcaddr& operator=(uint32 rhs) {
		value.l = rhs;
		return *this;
	}

	// 代入

	// アドレス、FC、R/W をセットする。fc_ は下位3ビットのみであること。
	void Set(uint32 addr_, fcattr attr_) {
		value.l = addr_;
		value.h = (uint32)attr_;
	}

	// 全体
	uint64 Get() const			{ return value.q; }
	uint32 GetAddr() const		{ return value.l; }
	uint32 GetAttr() const		{ return value.h; }
	void Set(uint64 q_)			{ value.q = q_; }
	void SetAddr(uint32 addr_)	{ value.l = addr_; }
	void SetAttr(uint32 attr_)	{ value.h = attr_; }

	uint64 GetSAddr() const		{ return value.q & 0x1'ffffffffULL; }
	uint64 GetFCAddr() const	{ return value.q & 0x7'ffffffffULL; }
	bool IsSuper() const		{ return (value.h & 0x01); }
	bool IsData()  const		{ return (value.h & 0x02); }
	bool IsWrite() const		{ return (value.h & 0x40) == 0; }
	uint32 GetFC() const		{ return (value.h >> 1) & 0x07; }
	uint32 GetSSW() const		{ return GetFC() | (value.h & 0x40); }

	void SetSuper(bool super_) {
		if (super_) {
			value.h |= 0x09;
		} else {
			value.h &= ~0x09;
		}
	}
	void SetData(bool data_) {
		value.h &= ~0x06;
		if (data_) {
			value.h |= (0x1 << 1);
		} else {
			value.h |= (0x2 << 1);
		}
	}
	void SetWrite(bool write_) {
		if (write_) {
			value.h &= ~0x40;
		} else {
			value.h |= 0x40;
		}
	}

	// (uint32)fcaddr でアドレスだけ取り出せると楽
	explicit operator uint32() const { return GetAddr(); }

	// fcaddr += n, -= n でアドレス進められると楽
	uint32 operator+=(uint32 val) {
		value.l += val;
		return value.l;
	}
	uint32 operator-=(uint32 val) {
		value.l -= val;
		return value.l;
	}
	// fcaddr++ でアドレス進められると楽
	fcaddr operator++(int n) {
		fcaddr old;
		old.Set(value.q);
		value.l++;
		return old;
	}
	// fcaddr &= mask でマスクが適用できると楽。
	fcaddr& operator&=(const uint32 mask_) {
		value.l &= mask_;
		return *this;
	}

 private:
	union64 value {};
};

// fcaddr と uint32 の AND 演算子。
// fcaddr & val だけでアドレスがマスクできて取り出せると楽。
static __unused uint32 operator&(fcaddr fcaddr_, uint32 rhs_) {
	return (uint32)fcaddr_ & rhs_;
}
