//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "mpu680x0.h"
#include "m68030mmu.h"
#include "mainbus.h"

#define TIA		(0)
#define TIB		(1)
#define TIC		(2)
#define TID		(3)
#define FCL		(4)	// 便宜上ね
#define MAX_LV	(5)

// ディスクリプタタイプ
#define GET_DT(x)	(x & 0x03)
#define DT_INVALID	(0)
#define DT_PAGE		(1)
#define DT_VALID4	(2)
#define DT_VALID8	(3)

// ディスクリプタ
#define MMU_DESC_S	(0x00000100)
#define MMU_DESC_CI	(0x00000040)
#define MMU_DESC_M	(0x00000010)
#define MMU_DESC_U	(0x00000008)
#define MMU_DESC_WP	(0x00000004)
#define MMU_DESC_DT	(0x00000003)

// エントリの状態
enum ATCtype {
	INVALID,
	NORMAL,
	EARLY,
	INDIRECT,
};

//
// デバッグ表示
//

// デバッグ表示のため値の表示に必要な桁数を返す
#define NIBBLE(x)		( (x) == 0 ? 1 : ((x) + 3) / 4 )

// MMU デバッグモードなら ATC エントリ a を表示
#define FILL_PRINT(fc, a) cpu->putlogf(3, [&] { return (a)->fill_print(fc); })

// インデント付き string_format
static std::string indent_format(int indent, const char *fmt, ...)
	__printflike(2, 3);
static std::string
indent_format(int indent, const char *fmt, ...)
{
	va_list ap;
	char *buf;

	va_start(ap, fmt);
	vasprintf(&buf, fmt, ap);
	va_end(ap);

	std::string rv(indent, ' ');
	rv += buf;
	free(buf);

	return rv;
}

#define MMULOG(lv, indent, arg...) \
	gMPU->putlogf(lv, [&, lam_func = __func__] { \
		(void)lam_func; \
		return indent_format(indent, arg); \
	})

// テーブルサーチ操作クラス
class m68030MMU final
{
 public:
	m68030MMU(MPU680x0Device *);
	~m68030MMU();

	// サーチを実行
	void search(int level);

	// ATC にエントリを作成
	void make_atc(m68030ATCLine *);

	// ATC にエントリを作成 (デバッガアクセス用)
	uint64 make_atc_debugger();

 private:
	// リミットチェック
	bool limitcheck() const;

	// ディスクリプタを取得
	bool fetch_desc(uint32, int);

	// MMU による物理メモリ読み込み
	uint64 phys_read_32(uint32 addr);

	// MMU による物理メモリ書き込み
	uint64 phys_write_32(uint32 addr, uint32 data);

	// ディスクリプタを表示用に (デバッグ用)
	std::string sprintf_desc();

	// ディスクリプタタイプ文字列 (デバッグ用)
	const char *dtname(uint32 dt);

 public:
	ATCtype m_type {};		// 状態
	bool    m_s {};			// S ビット
	bool    m_wp {};		// WP ビット
	bool    m_ci {};		// CI ビット
	bool    m_m {};			// ページディスクリプタの M が立っているか
	bool    m_lim {};		// サーチ中にリミットを越えた
	bool	m_berr {};		// テーブルサーチ中にバスエラー
	bool	m_ptest {};		// PTEST[RW] 命令中なら真
	bool	m_debugger {};	// デバッガからのアクセスなら true にする
	int     x {};			// 段数 TIA〜TID, FCL
	uint32	m_lastaddr {};	// 最後にアクセスしたディスクリプタアドレス(PTEST)
 private:
	int     m_lv[MAX_LV] {};
	uint32	m_laddr {};
	int     m_fc {};
	uint32  m_desc1 {};		// ディスクリプタ
	uint32  m_desc2 {};		// ロングディスクリプタの後半部
	int     m_descsize {};	// ディスクリプタサイズ (4 or 8)
	int     m_lastsize {};	// 前段のディスクリプタサイズ

	// デバッグ用文字列
	static const char * const m_lvname[MAX_LV];

	MPU680x0Device *cpu;
};

// 現在の TC:E、TT:E の状態に応じてアドレス変換の有無を切り替える。
// mmu_enable ならアドレス変換ありのバスアクセスを使う。see m68030bus.cpp
void
MPU680x0Device::mmu_enable_changed()
{
	bool enable;

	// TC、TT0、TT1 のいずれかが有効ならアドレス変換あり
	enable = (mmu_tc.e || mmu_tt[0].e || mmu_tt[1].e);

	// 変化した時だけ
	if (mmu_enable && !enable) {
		// 無効にする
		mmu_enable = false;
		putlog(1, "MMU Translation Disabled");
	} else if (!mmu_enable && enable) {
		// 有効にする
		mmu_enable = true;
		putlog(1, "MMU Translation Enabled");
	}
}


//
// レジスタクラス
//

bool
MPU680x0Device::SetSRP(uint32 h, uint32 l)
{
	if ((h & m68030RP::DT_MASK) == 0) {
		return false;
	}
	reg.srp.h = h & m68030RP::H_MASK;
	reg.srp.l = l & m68030RP::L_MASK;
	return true;
}

bool
MPU680x0Device::SetCRP(uint32 h, uint32 l)
{
	if ((h & m68030RP::DT_MASK) == 0) {
		return false;
	}
	reg.crp.h = h & m68030RP::H_MASK;
	reg.crp.l = l & m68030RP::L_MASK;
	return true;
}

void
MPU680x0Device::SetTT(int n, uint32 data)
{
	uint32 lbase;
	uint32 lmask;
	uint32 sbase;
	uint32 smask;

	// マスクしてレジスタイメージにセット
	reg.tt[n] = data & m68030TT::MASK;

	// data を分解してメンバにセット。
	//
	//    3                   2                   1                   0
	//  1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
	// +---------------+---------------+-+-+-+-+-+-+-+-+-+-----+-+-----+
	// |LogicalAddrBase|LogicalAddrMask|E|       |C|R|M| | FCB | | FCM |
	// +---------------+---------------+-+-+-+-+-+-+-+-+-+-----+-+-----+

	mmu_tt[n].e      = (data & m68030TT::E);
	mmu_tt[n].ci     = (data & m68030TT::CI);

	// 下位ロングワード
	// ・lbase は比較先アドレスで、ff000000の位置。
	// ・lmask はアドレスマスク。%1 が無視を表しているので NOT しておく。
	//
	//  target  AAAAAAAA xxxxxxxx xxxxxxxx xxxxxxxx : Aは論理アドレス
	//        ^                                     : xは不問
	//  lbase   BBBBBBBB 00000000 00000000 00000000 : Bは論理アドレスベース
	//        &
	//  lmask   MMMMMMMM 00000000 00000000 00000000 : Mは論理アドレスマスク
	//
	lbase  = (data & m68030TT::LBASE_MASK);
	lmask  = (~data & m68030TT::LMASK_MASK) << 8;

	// 上位ロングワード
	// ・E ビットは位置はどこでもいいので TT:E ビットの
	//   位置をそのまま使用する。target 側のこの位置は必ずビットが落ちている
	//   ため base 側にビットを立てておくと必ず不一致になることを利用して、
	//   E ビットの評価も一度に行う。
	// ・RW ビットは SSW の RW と比較するので、SSW_RW のビット位置を使う
	//   (0x00000040 の位置)。TT:RW も SSW_RW も Read が %1。
	// ・RWM が %1 なら R/W は無視、RWM が %0 なら R/W の一致も検査するので
	//   RWM が %0 の時は RW のマスクビットを立てておく。
	// ・FC は fcattr の配置に合わせて1ビット左シフトしたところを使う。
	// ・FCMask は %1 が無視を表しているので NOT しておく。
	//
	//  target  00000000 00000000 00000000 0R00FFFS : R はアクセス R/~W
	//        ^                                     : F はアクセス FC
	//                                              : S は FC2 のコピー
	//  base    00000000 00000000 E0000000 0R00FFF0 : R は期待する R/~W
	//        &                                     : F は期待する FC
	//                                              : E は ~TT:E
	//  mask    00000000 00000000 10000000 0M00CCC0 : M は R/~W マスク
	//                                              : C は FC マスク
	sbase  = ~data & m68030TT::E;
	sbase |= (data & m68030TT::RW) >> 3;
	sbase |= (data & m68030TT::FCBASE_MASK) >> 3;
	smask  = m68030TT::E;
	if ((data & m68030TT::RWM) == 0) {
		smask |= M68K::SSW_RW;
	}
	smask |= (~data & m68030TT::FCMASK_MASK) << 1;

	mmu_tt[n].base = lbase | (((uint64)sbase) << 32);
	mmu_tt[n].mask = lmask | (((uint64)smask) << 32);

	// 論理アドレスとEnableビットだけの早見表を作成。
	// TT0 が一致するなら bit0 をセット、TT1 が一致するなら bit1 をセット
	// しておく。
	for (int i = 0; i < countof(mmu_tt_quick); i++) {
		uint32 laddr = ((uint32)i) << 24;
		if (mmu_tt[n].e && ((laddr ^ lbase) & lmask) == 0) {
			mmu_tt_quick[i] |= (1 << n);
		} else {
			mmu_tt_quick[i] &= ~(1 << n);
		}
	}

	mmu_enable_changed();
}

// アクセスがこの TT とマッチするか調べる。
// laddr の上位ワードはそのまま使える。
// 下位32ビットが論理アドレス (下24ビットは呼び出し側でマスクしなくてよい)。
// この TT が有効でアドレスがマッチすれば true を返す。
// (TT:E は演算に織り込んである、このすぐ上の SetTT() のコメント参照)
bool
m68030TT::Match(fcaddr laddr) const
{
	bool rv = (((laddr.Get() ^ base) & mask) == 0);
	return rv;
}

// 成功すれば true、MMU 構成例外が起きる場合は false を返す。
// 呼び出し側で例外を起こすこと。
bool
MPU680x0Device::SetTC(uint32 data)
{
	// マスクしてレジスタイメージにセット
	reg.tc = data & m68030TC::MASK;

	// data を分解してメンバにセット
	mmu_tc.Set(data);

	// 有効にする場合は一貫性を確認
	if (mmu_tc.e) {
		if (mmu_tc.Check() == false) {
			// エラーなら E をクリアして MMU 構成例外。
			// (例外処理自体は呼び出し側で行う)
			mmu_tc.e = false;
			mmu_enable_changed();
			return false;
		}
	}

	// マスクをあらかじめ用意しておく
	mmu_tc.MakeMask();

	// TC:E が変更されていればスイッチ更新
	mmu_enable_changed();
	return true;

}

// TC レジスタへの書き込み値を m68030TC クラスにセットする。
void
m68030TC::Set(uint32 data)
{
	e   = data & TC_E;
	sre = data & TC_SRE;
	fcl = data & TC_FCL;
	tid = (data)       & 0x0f;
	tic = (data >>= 4) & 0x0f;
	tib = (data >>= 4) & 0x0f;
	tia = (data >>= 4) & 0x0f;
	is  = (data >>= 4) & 0x0f;
	ps  = (data >>= 4) & 0x0f;
}

// 現在の値が MMU 構成例外になるなら false を返す。
bool
m68030TC::Check() const
{
	if (ps < 8)
		return false;
	if (tia == 0)
		return false;
	if (tib == 0 && (tic > 0 || tid > 0))
		return false;
	if (tic == 0 && tid > 0)
		return false;
	if (ps + is + tia + tib + tic + tid != 32)
		return false;

	return true;
}

// 論理マスク、ページマスクを作成。(テーブルサーチで参照される)
void
m68030TC::MakeMask()
{
	// lmask (論理マスク) は TIA〜TID 部分のマスク。
	// pgmask (ページマスク) は PS 部分のマスク。
	//
	// 例えば IS=$8, TIA=$3, TIB=$4, TIC=$4, TID=$0, PS=$d
	// (X68030 IPL 内の MMU 判定のケース) なら
	//
	//           3          2          1          0
	//          10987654 32109876 54321098 76543210
	//         +--------+--------+--------+--------+
	//         |IIIIIIII AAABBBBC CCCPPPPP PPPPPPPP|
	//         +--------+--------+--------+--------+
	// lmask  = 00000000 11111111 11100000 00000000
	// pgmask = 00000000 00000000 00011111 11111111
	lmask = (1 << (tia + tib + tic + tid)) - 1;
	lmask <<= ps;
	pgmask = (1 << ps) - 1;

	// アドレスを (IS,) TIA 〜 TID に分解。
	//
	// TIA=$C, TIB=$A, (TIC = TID = 0), PS=$A の場合以下のようにする。
	// 使ってない TIC, TID については tix[TIC] == 0 で判断する。
	//
	//                    A              B             PS
	//           +----------------+--------------+--------------+
	// $00A01A00 | 0000 0000 1010 | 0000 0001 10 | xx xxxx xxxx |
	//           +----------------+--------------+--------------+
	//            shift[TIA]=20     shift[TIB]=10
	//            mask[TIA]=0xfff   mask[TIB]=0x3ff
	int sh = ps;
	for (int i = TID; i >= TIA; i--) {
		if (tix[i] > 0) {
			shift[i] = sh;
			mask[i] = (1 << tix[i]) - 1;
			sh += tix[i];
		}
	}
}

void
MPU680x0Device::SetMMUSR(uint16 data)
{
	reg.mmusr = data & m68030MMUSR::MASK;
}


//
// ATC テーブル
//

// コンストラクタ
m68030ATCTable::m68030ATCTable()
{
#if !defined(ATC_SINGLE)
	for (int i = 0; i < countof(line); i++) {
		line[i].Invalidate();
	}
	head = &line[0];
#else
	for (int i = 0; i < countof(line); i++) {
		line[i].Invalidate();
		insert_tail(&line[i]);
	}
#endif
}

// デストラクタ
m68030ATCTable::~m68030ATCTable()
{
}

#if !defined(ATC_SINGLE)
// a のエントリを先頭にする。
m68030ATCLine *
m68030ATCTable::MoveHead(m68030ATCLine *a)
{
	// head を向け直す
	head = a;

	// 更新
	head->age = counter++;

	// そのエントリを返す
	return head;
}
#endif


//
// ATC
//

// コンストラクタ
m68030ATC::m68030ATC(MPU680x0Device *cpu_)
{
	cpu = cpu_;
}

// デストラクタ
m68030ATC::~m68030ATC()
{
}

#if defined(ATC_SINGLE)
// item をリストの先頭に挿入。
void
m68030ATCTable::insert_head(m68030ATCLine *item)
{
	item->prev = NULL;
	item->next = head;
	if (head) {
		head->prev = item;
	}
	head = item;
}

// item をリストの先頭から2番目に挿入。
void
m68030ATCTable::insert_second(m68030ATCLine *item)
{
	item->prev = head;
	if (head) {
		item->next = head->next;
		if (head->next) {
			head->next->prev = item;
		}
		head->next = item;
	} else {
		item->next = NULL;
	}
}

// item をリストの末尾に追加。
void
m68030ATCTable::insert_tail(m68030ATCLine *item)
{
	item->next = NULL;
	if (head == NULL) {
		item->prev = NULL;
		head = item;
	} else {
		m68030ATCLine *a = head;
		while (a->next)
			a = a->next;
		a->next = item;
		item->prev = a;
	}
}

// item をリストから外す
void
m68030ATCTable::remove(m68030ATCLine *item)
{
	if (item->prev)
		item->prev->next = item->next;
	if (item->next)
		item->next->prev = item->prev;
	if (item == head)
		head = item->next;
}

// item をリストの先頭に移す。
// fill_fetch() から呼ばれるもので、リストは空でなく、item は先頭から
// 3番目以内ではない、という前提で高速化。
void
m68030ATCTable::move_head(m68030ATCLine *item)
{
	// 先頭でない前提の remove
	item->prev->next = item->next;
	if (item->next)
		item->next->prev = item->prev;

	// 先頭がある前提の insert_head
	item->prev = NULL;
	item->next = head;
	head->prev = item;
	head = item;
}

// item を先頭から2番目に移す。
// fill_{read,write} から呼ばれるもので、リストは空でなく、item は先頭から
// 3番目以内ではない、という前提で高速化。
void
m68030ATCTable::move_second(m68030ATCLine *item)
{
	// 先頭でない前提の remove
	item->prev->next = item->next;
	if (item->next)
		item->next->prev = item->prev;

	// 先頭2つがある前提の insert_second
	item->prev = head;
	item->next = head->next;
	head->next->prev = item;
	head->next = item;
}
#endif // ATC_SINGLE

// fill_{fetch,read,write}() が各アクセス用の ATC サーチ。
// あらかじめ論理ページにマスクされたアドレス fc:addr が ATC にあれば
// そのエントリを返す。なければ ATC を更新してそのエントリを返す。
// fill_peek() はデバッグ用なので一切状態を更新せず結果だけ返す。

const m68030ATCLine&
m68030ATC::fill_fetch(fcaddr fcaddr)
{
	uint32 laddr = fcaddr.GetAddr();
	uint fc = fcaddr.GetFC();
	m68030ATCTable *table;
	m68030ATCLine *a;
	int i;

#if !defined(ATC_SINGLE)
	table = &tables[fc];
	// キャッシュとマッチすればそれを返す
	a = table->head;
	if (a->GetTag() == laddr) {
		FILL_PRINT(fc, a);
		table->hit_head++;
		return *a;
	}
	for (i = 0; i < countof(table->line); i++) {
		a = &table->line[i];
		if (a->GetTag() == laddr) {
			// ヒットしたので先頭に移動。fetch は常に入れ替え
			a = table->MoveHead(a);

			FILL_PRINT(fc, a);
			table->hit[i]++;
			return *a;
		}
	}

	// 見付からなかったので、古いエントリを1つ更新。
	a = table->MoveHead(table->Lookup());
	table->hit[m68030ATCTable::LINES]++;
	// タグはこっちで初期化
	a->tag = laddr;
#else
	table = &tables[0];
	// キャッシュとマッチすればそれを返す
	a = table->head;
	for (i = 0; a; a = a->next, i++) {
		if (a->GetTag() == laddr && a->fc == fc) {
			// ヒットしたので先頭に移動。
			if (i > 2) {
				table->move_head(a);
			}

			FILL_PRINT(fc, a);
			table->hit[i]++;
			return *a;
		}
	}

	// 見付からなかったので、古いエントリを1つ更新。
	a = table->lookup();
	table->remove(a);
	table->insert_head(a);
	table->hit[m68030ATCTable::LINES]++;
	// タグはこっちで初期化
	a->tag = laddr;
	a->fc  = fc;
#endif

	// テーブルサーチして結果を ATC に入れる
	m68030MMU mmu(cpu);
	mmu.search(7);
	mmu.make_atc(a);

	FILL_PRINT(fc, a);
	return *a;
}

const m68030ATCLine&
m68030ATC::fill_read(fcaddr fcaddr)
{
	uint32 laddr = fcaddr.GetAddr();
	uint fc = fcaddr.GetFC();
	m68030ATCTable *table;
	m68030ATCLine *a;
	int i;

#if !defined(ATC_SINGLE)
	table = &tables[fc];
	// キャッシュとマッチすればそれを返す
	a = table->head;
	if (a->GetTag() == laddr) {
		FILL_PRINT(fc, a);
		table->hit_head++;
		return *a;
	}
	for (i = 0; i < countof(table->line); i++) {
		a = &table->line[i];
		if (a->GetTag() == laddr) {
			// ヒットしたので先頭に移動。
			a = table->MoveHead(a);

			FILL_PRINT(fc, a);
			table->hit[i]++;
			return *a;
		}
	}

	// 見付からなかったので、古いエントリを1つ更新
	a = table->MoveHead(table->Lookup());
	table->hit[m68030ATCTable::LINES]++;
	// タグはこっちで初期化
	a->tag = laddr;
#else
	table = &tables[0];
	// キャッシュとマッチすればそれを返す
	a = table->head;
	for (i = 0; a; a = a->next, i++) {
		if ((a->GetTag() == laddr) && (a->fc == fc)) {
			// ヒットしたので先頭から2番目に移動
			if (i > 2) {
				table->move_second(a);
			}

			FILL_PRINT(fc, a);
			table->hit[i]++;
			return *a;
		}
	}

	// 見付からなかったので、古いエントリを1つ更新
	a = table->lookup();
	table->remove(a);
	table->insert_head(a);
	table->hit[m68030ATCTable::LINES]++;
	// タグはこっちで初期化
	a->tag = laddr;
	a->fc  = fc;
#endif

	// テーブルサーチして結果を ATC に入れる
	m68030MMU mmu(cpu);
	mmu.search(7);
	mmu.make_atc(a);

	FILL_PRINT(fc, a);
	return *a;
}

const m68030ATCLine&
m68030ATC::fill_write(fcaddr fcaddr)
{
	uint32 laddr = fcaddr.GetAddr();
	uint fc = fcaddr.GetFC();
	m68030ATCTable *table;
	m68030ATCLine *a;
	int i;

#if !defined(ATC_SINGLE)
	table = &tables[fc];
	// キャッシュとマッチすればそれを返す
	a = table->head;
	if (a->GetTag() == laddr) {
		// ライトアクセスなので、WP でなく Modified == 0 なら
		// Modified を立てるため、このエントリを破棄して改めて
		// テーブルサーチを実行する。
		if (!a->IsWProtect() && !a->IsModified()) {
			FILL_PRINT(fc, a);
			MMULOG(3, 1, "%s: write access but M is not set, "
					"so invalidate it", lam_func);
			a->Invalidate();
			goto search;
		}
		FILL_PRINT(fc, a);
		table->hit_head++;
		return *a;
	}
	for (i = 0; i < countof(table->line); i++) {
		a = &table->line[i];
		if (a->GetTag() == laddr) {
			// ライトアクセスなので、WP でなく Modified == 0 なら
			// Modified を立てるため、このエントリを破棄して改めて
			// テーブルサーチを実行する。
			if (!a->IsWProtect() && !a->IsModified()) {
				FILL_PRINT(fc, a);
				MMULOG(3, 1, "%s: write access but M is not set, "
						"so invalidate it", lam_func);
				a->Invalidate();
				break;
			}
			// ヒットしたので先頭に移動。
			a = table->MoveHead(a);

			FILL_PRINT(fc, a);
			table->hit[i]++;
			return *a;
		}
	}
 search:;

	// 見付からなかったので、古いエントリを1つ更新
	a = table->MoveHead(table->Lookup());
	table->hit[m68030ATCTable::LINES]++;
	// タグはこっちで初期化
	a->tag = laddr;
#else
	table = &tables[0];
	// キャッシュとマッチすればそれを返す
	a = table->head;
	for (i = 0; a; a = a->next, i++) {
		if ((a->GetTag() == laddr) && (a->fc == fc)) {
			// ライトアクセスなので、WP でなく Modified == 0 なら
			// Modified を立てるため、このエントリを破棄して改めて
			// テーブルサーチを実行する。
			if (!a->IsWProtect() && !a->IsModified()) {
				FILL_PRINT(fc, a);
				MMULOG(3, 1, "%s: write access but M is not set, "
						"so invalidate it", lam_func);
				a->Invalidate();
				break;
			}

			// ヒットしたので、先頭から2番目に移動
			if (i > 2) {
				table->move_second(a);
			}

			FILL_PRINT(fc, a);
			table->hit[i]++;
			return *a;
		}
	}

	// 見付からなかったので、古いエントリを1つ更新
	a = table->lookup();
	table->remove(a);
	table->insert_head(a);
	// 論理部分(タグ部分)を初期化
	a->tag = laddr;
	a->fc  = fc;
	table->hit[m68030ATCTable::LINES]++;
#endif

	// テーブルサーチして結果を ATC に入れる
	m68030MMU mmu(cpu);
	mmu.search(7);
	mmu.make_atc(a);

	FILL_PRINT(fc, a);
	return *a;
}

// ピークアクセス用に ATC を検索する。
// ATC になければ NULL を返す。ここでは疑似テーブルサーチは行わない。
// laddr は論理ページアドレスにマスクしておくこと。
const m68030ATCLine *
m68030ATC::fill_peek(uint32 laddr, uint fc)
{
	m68030ATCTable *table;
	const m68030ATCLine *a;

#if !defined(ATC_SINGLE)
	// キャッシュとマッチすればそれを返す
	table = &tables[fc];
	for (int i = 0; i < countof(table->line); i++) {
		a = &table->line[i];
		if (a->GetTag() == laddr) {
			// ヒットした
			MMULOG(3, 1, "%s: match %d:$%08x -> $%08x", lam_func,
				fc, laddr, (uint32)a->GetPAddr());
			return a;
		}
	}
#else
	// キャッシュとマッチすればそれを返す
	table = &tables[0];
	for (a = table->head; a; a = a->next) {
		if ((a->GetTag() == laddr) && ((a->fc & 4) == fc)) {
			// ヒットした
			MMULOG(3, 1, "%s: match %d:$%08x -> $%08x", lam_func,
				a->fc, laddr, (uint32)a->GetPAddr());
			return a;
		}
	}
#endif

	return NULL;
}

// デバッグ表示。
// ATC_SINGLE なら fc は ATCLine に含まれているため、引数はダミーで
// 変数名が衝突するため仮引数名も変えてある。なんだかなあ。
// XXX print といいつつ string を返す
#if defined(ATC_SINGLE)
std::string
m68030ATCLine::fill_print(uint dummyfc)
#else
std::string
m68030ATCLine::fill_print(uint fc)
#endif
{
	uint32 laddr = GetLAddr();

	if (!IsBusError()) {
		uint32 ci    = IsCInhibit();
		uint32 wp    = IsWProtect();
		uint32 mod   = IsModified();
		return indent_format(1,
			"ATC_fill: match %d:$%08x -> $%08x CI=%d WP=%d M=%d",
			fc, laddr, GetPAddr(), ci, wp, mod);
	} else {
		return indent_format(1,
			"ATC_fill: match %d:$%08x -> BusErr",
			fc, laddr);
	}
}

#if !defined(ATC_SINGLE)
// 更新用にエントリを1つ差し出す。
m68030ATCLine *
m68030ATCTable::Lookup()
{
	uint32 minage;
	int minidx;

	// 空きエントリを探しながら、最古エントリを探しておく。
	// 最古エントリは age が最も小さいもの。
	minage = (uint32)-1;
	minidx = 0;
	for (int i = 0; i < countof(line); i++) {
		m68030ATCLine& a = line[i];
		if (a.IsInvalid()) {
			MMULOG(3, 1, "ATC.Lookup got free entry [%d]", i);
			return &a;
		}
		if (a.age < minage) {
			minage = a.age;
			minidx = i;
		}
	}
	// 空きがなければ、最古のエントリを差し出す
	MMULOG(3, 1, "ATC.Lookup got oldest entry [%d]", minidx);
	return &line[minidx];
}
#else
// 更新用にエントリを1つ差し出す。
m68030ATCLine *
m68030ATCTable::lookup()
{
	m68030ATCLine *a;

	// 空きエントリを探す
	for (int i = 0; i < LINES; i++) {
		a = &line[i];
		if (a->IsInvalid()) {
			MMULOG(3, 1, "ATC.Lookup got free entry [%d]", i);
			goto found;
		}
	}

	// 空きがなければ、リストの末尾が最古なのでこれを差し出す
	a = head;
	while (a->next) {
		a = a->next;
	}
	MMULOG(3, 1, "ATC.Lookup got oldest entry");

 found:
	return a;
}
#endif

// ATC をすべてフラッシュする。命令からコールされる
void
m68030ATC::flush()
{
	for (int i = 0; i < countof(tables); i++) {
		tables[i].flush();
	}
}

// ATC の fc, mask が一致するエントリをフラッシュする。
// 命令からコールされる。
void
m68030ATC::flush(uint32 fc, uint32 mask)
{
#if !defined(ATC_SINGLE)
	for (int i = 0; i < countof(tables); i++) {
		if (((i ^ fc) & mask) == 0) {
			// FC が一致したテーブルを全フラッシュ
			tables[i].flush();
		}
	}
#else
	for (int i = 0; i < 8; i++) {
		if (((i ^ fc) & mask) == 0) {
			// この FC をテーブルからフラッシュ
			tables[0].flush_fc(i);
		}
	}
#endif
}

// ATC の fc, mask, addr が一致するエントリをフラッシュする。
// 命令からコールされる。
void
m68030ATC::flush(uint32 fc, uint32 mask, uint32 addr)
{
#if !defined(ATC_SINGLE)
	for (int i = 0; i < countof(tables); i++) {
		if (((i ^ fc) & mask) == 0) {
			// FC が一致したテーブルから addr が一致したエントリをフラッシュ
			tables[i].flush(addr);
		}
	}
#else
	for (int i = 0; i < 8; i++) {
		if (((i ^ fc) & mask) == 0) {
			// FC と addr が一致したエントリをフラッシュ
			tables[0].flush(i, addr);
		}
	}
#endif
}

// この ATC テーブルのエントリをすべてフラッシュする。
void
m68030ATCTable::flush()
{
	for (int i = 0; i < countof(line); i++) {
		m68030ATCLine& a = line[i];
		a.Invalidate();
	}
}

#if !defined(ATC_SINGLE)
// addr が一致するエントリをすべてフラッシュ
void
m68030ATCTable::flush(uint32 addr)
{
	for (int i = 0; i < countof(line); i++) {
		m68030ATCLine& a = line[i];
		if (a.GetLAddr() == addr) {
			a.Invalidate();
		}
	}
}
#else
// FC が一致するエントリをすべてフラッシュする
void
m68030ATCTable::flush_fc(uint32 fc)
{
	for (int i = 0; i < countof(line); i++) {
		m68030ATCLine& a = line[i];
		if (a.fc == fc) {
			a.Invalidate();
		}
	}
}

// FC と addr が一致するエントリをすべてフラッシュ
void
m68030ATCTable::flush(uint32 fc, uint32 addr)
{
	for (int i = 0; i < countof(line); i++) {
		m68030ATCLine& a = line[i];
		if (a.fc == fc && a.GetLAddr() == addr) {
			a.Invalidate();
		}
	}
}
#endif


//
// テーブルサーチ
//

// コンストラクタ
m68030MMU::m68030MMU(MPU680x0Device *cpu_)
{
	cpu = cpu_;

	m_type = INVALID;
	x = -1;
	m_fc = -1;
}

// デストラクタ
m68030MMU::~m68030MMU()
{
}

// テーブルサーチのメイン部分。
// m_type および ATC 作成に必要な情報を埋めてから帰る。
// ATC は帰った先で作成する。
void
m68030MMU::search(int level)
{
	const char *rpname;
	int dt;
	bool is_super;

	m_laddr = cpu->bus.laddr.GetAddr();
	m_fc    = cpu->bus.GetFC();

	is_super = (m_fc & 0x04);

	// アドレスを (IS,) TIA .. TID に分解。
	//
	// TIA=$C, TIB=$A, (TIC = TID = 0), PS=$A の場合以下のようにする。
	// 使ってない TIC, TID については cpu->mmu_tc->{tic,tid} == 0 で判断する。
	//
	//                    A              B             PS
	//           +----------------+--------------+--------------+
	// $00A01A00 | 0000 0000 1010 | 0000 0001 10 | xx xxxx xxxx |
	//           +----------------+--------------+--------------+
	//            shift[TIA] =20   shift[TIB] =10
	//            lv[TIA].idx=0xA  lv[TIB].idx=0x6
	//
	for (int i = TID; i >= TIA; i--) {
		if (cpu->mmu_tc.tix[i] > 0) {
			uint32 mask = cpu->mmu_tc.mask[i];
			m_lv[i] = (m_laddr >> cpu->mmu_tc.shift[i]) & mask;
		}
	}
	m_lv[FCL] = m_fc;

	cpu->putlogf(3, [&] {
		std::string buf;

		buf = indent_format(2, "search: laddr=$%08x ->", m_laddr);
		if (cpu->mmu_tc.is > 0)
			buf += string_format(" IS=%d", cpu->mmu_tc.is);
		if (cpu->mmu_tc.fcl)
			buf += string_format(" FC=%d", m_lv[FCL]);
		for (int i = TIA; i <= TID; i++) {
			if (cpu->mmu_tc.tix[i] > 0) {
				buf += string_format(" %s=%d($%0*x)",
					m_lvname[i], cpu->mmu_tc.tix[i],
					NIBBLE(cpu->mmu_tc.tix[i]), m_lv[i]);
			}
		}
		return buf;
	});

	// この辺に図9-26 の初期化だがコンストラクタで初期化済み。

	// ここから図9-25。

	// 使用するルートポインタを決定
	if (cpu->mmu_tc.sre && is_super) {
		rpname = "SRP";
		m_desc1 = cpu->GetSRPh();
		m_desc2 = cpu->GetSRPl();
	} else {
		rpname = "CRP";
		m_desc1 = cpu->GetCRPh();
		m_desc2 = cpu->GetCRPl();
	}

	// ルートポインタの DT をチェック
	dt = GET_DT(m_desc1);
	MMULOG(3, 2, "%s: %s dt=%s", lam_func, rpname, dtname(dt));
	switch (dt) {
	 case DT_INVALID:
		// 無効は設定できないはずなので、ここに来ないはず
		PANIC("%s.DT==Invalid", rpname);
		break;

	 case DT_PAGE:		// ページディスクリプタ
		m_type = EARLY;
		return;

	 case DT_VALID4:	// 有効(4バイト)
	 case DT_VALID8:	// 有効(8バイト)
		// descsize は DT=$2 なら 4、DT=$3 なら 8 になる
		m_descsize = 1 << dt;
		m_lastsize = 8;
		break;
	}

	// 必要ならファンクションコード・ルックアップ
	if (cpu->mmu_tc.fcl) {
		x = FCL;
		// たぶん検索レベルを1つ消費するはず。未確認。
		level--;

		MMULOG(3, 2, "%s: %s tableaddr=$%08x idx=$%x", lam_func,
			m_lvname[x], m_desc2, m_lv[x]);

		// ディスクリプタをフェッチ
		if (fetch_desc(m_desc2, m_lv[x]) == false) {
			return;
		}

		// ディスクリプタタイプをチェック
		dt = GET_DT(m_desc1);
		switch (dt) {
		 case DT_INVALID:	// 無効
			m_type = INVALID;
			return;

		 case DT_PAGE:		// ページディスクリプタ
			m_type = EARLY;
			return;

		 case DT_VALID4:	// 有効4バイト
		 case DT_VALID8:	// 有効8バイト
			m_lastsize = m_descsize;
			m_descsize = 1 << dt;
			break;
		}
	}

	// TIx のテーブルサーチに入る。図9-25 の下半分。

	for (x = TIA; x < level; x++) {
		// リミットチェック
		if (m_lastsize == 8 && limitcheck() == false) {
			m_type = INVALID;
			m_lim = true;
			return;
		}

		// テーブルアドレス
		uint32 tableaddr;
		tableaddr = (m_lastsize == 4) ? m_desc1 : m_desc2;
		tableaddr &= 0xfffffff0;
		MMULOG(3, 2, "%s: %s tableaddr=$%08x idx=$%x", lam_func,
			m_lvname[x], tableaddr, m_lv[x]);

		// ディスクリプタをフェッチ
		if (fetch_desc(tableaddr, m_lv[x]) == false) {
			return;
		}

		// ディスクリプタタイプをチェック
		dt = GET_DT(m_desc1);
		switch (dt) {
		 case DT_INVALID:	// 無効
			m_type = INVALID;
			return;

		 case DT_PAGE:		// ページディスクリプタ
			if (x == TID) {
				m_type = NORMAL;
			} else {
				if (cpu->mmu_tc.tix[x + 1] == 0) {
					m_type = NORMAL;
				} else {
					m_type = EARLY;
				}
			}
			return;

		 case DT_VALID4:	// 有効4バイト
		 case DT_VALID8:	// 有効8バイト
			m_lastsize = m_descsize;
			m_descsize = 1 << dt;

			if (x == TID || cpu->mmu_tc.tix[x + 1] == 0) {
				// 間接
				m_type = INDIRECT;

				// ディスクリプタをフェッチ
				tableaddr = (m_lastsize == 4) ? m_desc1 : m_desc2;
				tableaddr &= 0xfffffff0;
				if (fetch_desc(tableaddr, 0) == false) {
					return;
				}
				if (GET_DT(m_desc1) != DT_PAGE) {
					m_type = INVALID;
				}
				return;
			}
			// 次段のサーチを繰り返す
			break;
		}
	}

	// レベル上限にひっかかった場合、x は過ぎているので一つ戻しておく。
	x--;
}

// リミットチェックを実行。図9-28。
// 無効のケースに落ちたら false を返す。
bool
m68030MMU::limitcheck() const
{
	uint16 limit = (m_desc1 >> 16) & 0x7fff;

	// 最上位ビットが L/U
	if ((int32)m_desc1 < 0) {
		if (m_lv[x] < limit) {
			MMULOG(3, 2, "%s: $%04x < $%04x", lam_func, m_lv[x], limit);
			return false;
		}
	} else {
		if (m_lv[x] > limit) {
			MMULOG(3, 2, "%s: $%04x > $%04x", lam_func, m_lv[x], limit);
			return false;
		}
	}
	return true;
}

// ディスクリプタ中のフラグの更新 (を予約する)
#define UPDATE_DESC(flag)	do { \
	if ((m_desc1 & (flag)) == 0) { \
		m_desc1 |= (flag); \
		do_write = true; \
	} \
} while (0)

// ディスクリプタのフェッチと、ヒストリおよびステータスの更新。
// 図9-29。
// ディスクリプタを取得できれば true、
// ATC エントリの作成に向かう場合は false。
bool
m68030MMU::fetch_desc(uint32 tableaddr, int idx)
{
	uint32 descaddr;
	bool do_write;
	uint dt;

	// ディスクリプタのアドレスを計算。ここだけマニュアルの説明が雑すぎ。
	// インダイレクトなら idx = 0 でコールしてあるので、
	// tableaddr がそのまま PA になるという寸法。
	descaddr = tableaddr + idx * m_descsize;
	MMULOG(3, 2, "%s: descsize=%d descaddr=$%08x", lam_func,
		m_descsize, descaddr);

	// バスエラー発生フラグをクリアしておく
	m_berr = false;

	// ディスクリプタを取得
	if (m_debugger) {
		// デバッガアクセス
		uint64 data;
		data = cpu->mainbus->Peek32(descaddr);
		if ((int64)data < 0) {
			goto buserr;
		}
		m_desc1 = data;

		if (m_descsize == 8) {
			data = cpu->mainbus->Peek32(descaddr + 4);
			if ((int64)data < 0) {
				goto buserr;
			}
			m_desc2 = data;
		}
	} else {
		uint64 data;

		data = phys_read_32(descaddr);
		if ((int64)data < 0) {
			goto buserr;
		}
		m_desc1 = data;

		if (m_descsize == 8) {
			data = phys_read_32(descaddr + 4);
			if ((int64)data < 0) {
				goto buserr;
			}
			m_desc2 = data;
		}

		m_lastaddr = descaddr;
	}

	// ディスクリプタタイプをチェック
	// DT は必ず最初のロングワードの下位2ビット。
	do_write = false;
	dt = GET_DT(m_desc1);
	MMULOG(3, 2, "%s: desc=%s dt=%s", lam_func,
		sprintf_desc().c_str(), dtname(dt));
	switch (dt) {
	 case DT_INVALID:	// 無効
		return true;

	 case DT_PAGE:		// ページディスクリプタ
		// U ビットを更新
		UPDATE_DESC(MMU_DESC_U);

		// 書き込み操作なら M を更新
		if (cpu->bus.IsWrite() && !m_ptest) {
			UPDATE_DESC(MMU_DESC_M);
		}
		// 図にはないけど、ATC 更新のためと PTEST のために
		// ページディスクリプタの M ビットを記録。
		if ((m_desc1 & MMU_DESC_M))
			m_m = true;
		break;

	 case DT_VALID4:	// 有効4バイト
	 case DT_VALID8:	// 有効8バイト
		// U ビットを更新
		UPDATE_DESC(MMU_DESC_U);
		break;
	}

	// デバッガ自身がアドレス変換を実施している時は書き込みは行わない
	if (m_debugger) {
		do_write = false;
	}

	// ライトライクルの実行
	if (do_write) {
		int64 berr = phys_write_32(descaddr, m_desc1);
		if (berr < 0) {
			goto buserr;
		}
		// バス動作正常終了
		MMULOG(3, 2, "%s: desc=%s updated", lam_func, sprintf_desc().c_str());
	}

	// ステータス更新 (OR していく)
	// XXX 実際には CI はページディスクリプタにしかないけど
	m_wp |= (m_desc1 & MMU_DESC_WP);
	m_ci |= (m_desc1 & MMU_DESC_CI);
	if (__predict_false(m_descsize == 8)) {
		m_s |= (m_desc1 & MMU_DESC_S);
	}

	return true;

	// バスエラー
 buserr:
	m_type = INVALID;
	m_berr = true;
	return false;
}

// ATC エントリを作成。実行系用。
// 図9-27 だが雑すぎ。
void
m68030MMU::make_atc(m68030ATCLine *a)
{
	std::string buf;
	uint32 paddr;

	a->data = 0;
	switch (m_type) {
	 case INVALID:	// 無効
		a->data |= m68030ATCLine::BUSERROR;
		MMULOG(3, 2, "%s: type=INVALID -> BusErr", lam_func);
		return;

	 case NORMAL:	// ノーマル
	 case INDIRECT:	// 間接
		paddr = (m_descsize == 4) ? m_desc1 : m_desc2;
		MMULOG(3, 2, "%s: type=NORMAL paddr=$%08x", lam_func,
			paddr & 0xffffff00);
		break;

	 case EARLY:	// アーリーターミネーション
		paddr = (m_descsize == 4) ? m_desc1 : m_desc2;
		buf = string_format("type=EARLY paddr=$%08x", paddr & 0xffffff00);

		// 残ってるビットも足して返す
		for (int i = x + 1; i <= TID; i++) {
			if (cpu->mmu_tc.tix[i] > 0) {
				paddr |= m_lv[i] << cpu->mmu_tc.shift[i];
				buf += string_format(" +%s=%08x", m_lvname[i], paddr);
			}
		}
		MMULOG(3, 2, "%s: %s", lam_func, buf.c_str());
		break;
	 default:
		PANIC("corrupted m_type=%d", (int)m_type);
	}

	// PS の部分は反映しない
	a->paddr = paddr & cpu->mmu_tc.lmask;

	if (m_ci)
		a->data |= m68030ATCLine::CINHIBIT;
	if (m_wp)
		a->data |= m68030ATCLine::WPROTECT;
	if (m_m)
		a->data |= m68030ATCLine::MODIFIED;
}

// make_atc() と同様にテーブルサーチの結果から物理ページアドレスを返す。
// バスエラーになるケースなら (uint64)-1 を返す。
// デバッガアクセスなので、VM に一切干渉しない。
uint64
m68030MMU::make_atc_debugger()
{
	uint64 paddr;

	switch (m_type) {
	 case INVALID:	// 無効
		return (uint64)-1;
		break;

	 case NORMAL:	// ノーマル
	 case INDIRECT:	// 間接
		paddr = (m_descsize == 4) ? m_desc1 : m_desc2;
		paddr &= 0xffffff00;
		break;

	 case EARLY:	// アーリーターミネーション
		paddr = (m_descsize == 4) ? m_desc1 : m_desc2;
		paddr &= 0xffffff00;
		// 残ってるビットも足して返す
		for (int i = x + 1; i <= TID; i++) {
			if (cpu->mmu_tc.tix[i] > 0) {
				paddr |= m_lv[i] << cpu->mmu_tc.shift[i];
			}
		}
		break;
	 default:
		PANIC("corrupted m_type=%d", (int)m_type);
	}
	return paddr;
}

// 物理メモリからの読み込み。
// 正常に読み込めれば 32bit の値。バスエラーなら (uint64)-1 を返す。
uint64
m68030MMU::phys_read_32(uint32 addr)
{
	m68kbus backup;
	uint64 data;

	// バス情報を一旦退避
	backup = cpu->bus;

	// 物理バスアクセス
	cpu->bus.laddr.Set(addr, fcattr::S | fcattr::D | fcattr::R);
	cpu->bus.paddr = cpu->bus.laddr;
	data = cpu->mainbus->Read32(cpu->bus.paddr);
	if (__predict_false((int64)data < 0)) {
		cpu->bus.size  = M68K::SSW_SIZE_LONG;
		return data;
	}

	// バス情報を復元
	cpu->bus = backup;
	return data;
}

// 物理メモリへの書き込み。
// 正常に書き込めれば 0、バスエラーなら (uint64)-1 を返す。
uint64
m68030MMU::phys_write_32(uint32 addr, uint32 data)
{
	m68kbus backup;
	uint64 berr;

	// バス情報を一旦退避
	backup = cpu->bus;

	// 物理バスアクセス
	cpu->bus.laddr.Set(addr, fcattr::S | fcattr::D | fcattr::W);
	cpu->bus.paddr = cpu->bus.laddr;
	berr = cpu->mainbus->Write32(cpu->bus.paddr, data);
	if (__predict_false((int64)berr < 0)) {
		cpu->bus.size  = M68K::SSW_SIZE_LONG;
		return berr;
	}

	// バス情報を復元
	cpu->bus = backup;
	return 0;
}

// ディスクリプタ表示用の文字列を返す。
// m_descsize によって m_desc1, m_desc2 を適切に文字列にする。
std::string
m68030MMU::sprintf_desc()
{
	std::string buf;

	if (m_descsize == 4) {
		buf = string_format("%08x", m_desc1);
	} else {
		buf = string_format("%08x_%08x", m_desc1, m_desc2);
	}

	// フラグの出現条件がややこしいので注意
	int dt = GET_DT(m_desc1);
	buf += string_format(" (%s%s%s%c%c)",
		((m_descsize == 4) ? "x" : ((m_desc1 & MMU_DESC_S) ? "S" : "-")),
		((dt != DT_PAGE)   ? "x" : ((m_desc1 & MMU_DESC_CI) ? "C" : "-")),
		((dt != DT_PAGE)   ? "x" : ((m_desc1 & MMU_DESC_M)  ? "M" : "-")),
		(m_desc1 & MMU_DESC_U)  ? 'U' : '-',
		(m_desc1 & MMU_DESC_WP) ? 'W' : '-'
	);

	return buf;
}

// ディスクリプタタイプ文字列 (デバッグ用)
const char *
m68030MMU::dtname(uint32 dt)
{
	static char buf[32];
	switch (dt) {
	 case DT_INVALID:	return "invalid";
	 case DT_PAGE:		return "page";
	 case DT_VALID4:	return "valid4";
	 case DT_VALID8:	return "valid8";
	}
	sprintf(buf, "(invalid DT %d)", dt);
	return buf;
}

// デバッグ用文字列
const char * const m68030MMU::m_lvname[MAX_LV] = {
	"TIA",
	"TIB",
	"TIC",
	"TID",
	"FCL",	// この順ではないが、便宜上ここにおいてある
};

//
// アドレス変換
//

// 論理アドレスが TT0/TT1 どちらかと一致すれば (bus.ci を更新して) true を返す。
// 一致しなければ false を返す。
// 高速化のため mmu_tt_quick による評価は呼び出し側で行なっている。
bool
MPU680x0Device::CheckTT()
{
	bool ci;
	int match;

	match = 0;
	ci = false;
	for (int i = 0; i < 2; i++) {
		const m68030TT& tt = mmu_tt[i];
		if (tt.Match(bus.laddr)) {
			// 一致した
			match += i + 1;
			ci |= tt.ci != 0;
			MMULOG(3, 1, "%s: TT%d match", lam_func, i);
		}
	}

	// 一致したら PA <- LA
	if (match != 0) {
		MMULOG(3, 0, "%s: result %08x (match TT%s)", lam_func,
			(uint32)bus.laddr,
			(match == 3) ? "0 and TT1" : ((match == 1) ? "0" : "1"));
		bus.ci = ci;
		return true;
	}
	return false;
}

// フェッチ用のアドレス変換を行う。
// 変換できれば true、バスエラーなら false を返す。
bool
MPU680x0Device::TranslateFetch()
{
	fcaddr fcaddr = bus.laddr;
	uint32 laddr = (uint32)fcaddr;

	// 状態表示
	MMULOG(3, 0, "%s: enter  %c:$%08x pc=$%08x", lam_func,
		(fcaddr.IsSuper() ? 'S' : 'U'),
		laddr,
		ppc);

	// まず TT をチェック
	if (__predict_false(mmu_tt_quick[laddr >> 24] != 0)) {
		if (CheckTT()) {
			return true;
		}
	}

	// ATC からアドレスを取得
	if (__predict_true(mmu_tc.e)) {
		fcaddr.SetAddr(fcaddr & mmu_tc.lmask);
		const m68030ATCLine& a = atc.fill_fetch(fcaddr);
		if (__predict_false(a.IsBusError())) {
			MMULOG(3, 0, "%s: result BusErr (ATC:B)", lam_func);
			return false;
		}

		bus.paddr.SetAddr(a.GetPAddr() | (laddr & mmu_tc.pgmask));
		bus.ci = a.data & m68030ATCLine::CINHIBIT;
	}

	MMULOG(3, 0, "%s: result $%08x", lam_func, (uint32)bus.paddr);
	return true;
}

// リード用のアドレス変換を行う。
// 変換できれば true、バスエラーなら false を返す。
bool
MPU680x0Device::TranslateRead()
{
	fcaddr fcaddr = bus.laddr;
	uint32 laddr = (uint32)fcaddr;

	// 状態表示
	MMULOG(3, 0, "%s:  enter  %c:$%08x pc=%c:$%08x", lam_func,
		(fcaddr.IsSuper() ? 'S' : 'U'),
		laddr,
		IsSuper() ? 'S' : 'U',
		ppc);

	// まず TT をチェック
	if (__predict_false(mmu_tt_quick[laddr >> 24] != 0)) {
		if (CheckTT()) {
			return true;
		}
	}

	// ATC からアドレスを取得
	if (__predict_true(mmu_tc.e)) {
		fcaddr.SetAddr(fcaddr & mmu_tc.lmask);
		const m68030ATCLine& a = atc.fill_read(fcaddr);
		if (__predict_false(a.IsBusError())) {
			MMULOG(3, 0, "%s:  result BusErr (ATC:B)", lam_func);
			return false;
		}

		bus.paddr.SetAddr(a.GetPAddr() | (laddr & mmu_tc.pgmask));
		bus.ci = a.data & m68030ATCLine::CINHIBIT;
	}

	MMULOG(3, 0, "%s:  result $%08x", lam_func, (uint32)bus.paddr);
	return true;
}

// ライト用のアドレス変換を行う。
// 変換できれば true、バスエラーなら false を返す。
bool
MPU680x0Device::TranslateWrite()
{
	fcaddr fcaddr = bus.laddr;
	uint32 laddr = (uint32)fcaddr;

	// 状態表示
	MMULOG(3, 0, "%s: enter  %c:$%08x pc=%c:$%08x", lam_func,
		(fcaddr.IsSuper() ? 'S' : 'U'),
		laddr,
		IsSuper() ? 'S' : 'U',
		ppc);

	// まず TT をチェック
	if (__predict_false(mmu_tt_quick[laddr >> 24] != 0)) {
		if (CheckTT()) {
			return true;
		}
	}

	// ATC からアドレスを取得
	if (__predict_true(mmu_tc.e)) {
		fcaddr.SetAddr(fcaddr & mmu_tc.lmask);
		const m68030ATCLine& a = atc.fill_write(fcaddr);
		if (__predict_false(a.IsBusError())) {
			MMULOG(3, 0, "%s: result BusErr (ATC:B)", lam_func);
			return false;
		}
		if (__predict_false(a.IsWProtect())) {
			// Write アクセスで WP が立ってたらバスエラー
			MMULOG(3, 0, "%s: result Buserr (ATC:WP)", lam_func);
			return false;
		}

		bus.paddr.SetAddr(a.GetPAddr() | (laddr & mmu_tc.pgmask));
		bus.ci = a.data & m68030ATCLine::CINHIBIT;
	}

	MMULOG(3, 0, "%s: result $%08x", lam_func, (uint32)bus.paddr);
	return true;
}

// ピーク用のアドレス変換を行う。
// ATC にエントリが見付からない場合、do_search が true ならテーブル
// サーチまで行う。false ならテーブルサーチせずに帰る。
// 変換できれば対応する物理アドレスを、バスエラーなら (uint64)-1 を返す。
uint64
MPU680x0Device::TranslatePeek(fcaddr fcaddr, bool do_search)
{
	uint64 paddr;
	uint32 laddr = fcaddr.GetAddr();

	// 状態表示
	MMULOG(3, 0, "%s:  enter  %c:$%08x", lam_func,
		(fcaddr.IsSuper() ? 'S' : 'U'),
		laddr);

	// まず TT をチェック
	paddr = laddr;
	if (__predict_false(mmu_tt_quick[laddr >> 24] != 0)) {
		for (int i = 0; i < 2; i++) {
			m68030TT& tt = mmu_tt[i];
			if (tt.Match(fcaddr)) {
				// 一致した。
				// CI はここでは使わないため、一致したことだけわかればよい
				MMULOG(3, 0, "%s:  result $%08x (match tt)",
					lam_func, (uint32)paddr);
				return paddr;
			}
		}
	}

	// 次に TC (ATC と MMU テーブルサーチ)。
	if (mmu_tc.e) {
		// ATC からこっそりアドレスを取得
		const m68030ATCLine *a;
		a = atc.fill_peek(laddr & mmu_tc.lmask, fcaddr.GetFC());
		if (a != NULL) {
			paddr = a->GetPAddr();
			paddr |= laddr & mmu_tc.pgmask;
			MMULOG(3, 0, "%s:  result $%08x (match atc)",
				lam_func, (uint32)paddr);
			return paddr;
		}

		// 見付からなかったのでこっそりテーブルサーチ。
		if (do_search) {
			// mmu.search() はさすがに bus を使うので事前に退避。
			m68kbus backup = bus;

			bus.laddr = fcaddr;
			m68030MMU mmu(this);
			mmu.m_debugger = true;
			mmu.search(7);
			paddr = mmu.make_atc_debugger();
			// リストア
			bus = backup;

			if ((int64)paddr >= 0) {
				paddr |= laddr & mmu_tc.pgmask;
				MMULOG(3, 0, "%s:  result $%08x (table search)",
					lam_func, (uint32)paddr);
				return paddr;
			}
		}

		// TC が有効だがヒットしなかったらバスエラー
		MMULOG(3, 0, "%s:  result BusErr", lam_func);
		return (uint64)-1;
	}

	// TT が無効かまたは有効でもヒットせず、
	// かつ TC も無効だったら (こっちは「有効でヒットせず」は含まない)、
	// 論理アドレスをそのまま返す
	MMULOG(3, 0, "%s:  result $%08x (no translation)", lam_func, laddr);
	return laddr;
}

//
// 命令
//

// ir2 の下位5ビットから FC を返す。
// 不当パターンなら C++ の例外をスローする。
uint32
MPU680x0Device::get_fc()
{
	uint fcfield = ir2 & 0x1f;

	if (fcfield == 0) {
		return reg.GetSFC();
	} else if (fcfield == 1) {
		return reg.GetDFC();
	} else if (fcfield < 0x08) {
		// break;
	} else if (fcfield < 0x10) {
		return reg.D[fcfield & 7] & 7;
	} else if (fcfield < 0x18) {
		return fcfield & 7;
	}
	// 不当命令
	throw M68K::EXCEP_ILLEGAL;
}

// MMU 不当命令 (2ワード目で確定)
void
MPU680x0Device::mmu_op_illg2()
{
	putlog(1, "MMU illegal instruction %04x %04x", ir, ir2);
	Exception(M68K::EXCEP_FLINE);
}

// PFLUSH fc,#mask 命令 (EA を持たない方)
void
MPU680x0Device::mmu_op_pflush()
{
	uint32 mask;
	uint32 fc;

	mask = (ir2 >> 5) & 7;
	fc = get_fc();
	atc.flush(fc, mask);
}

// PFLUSH fc,#mask,ea 命令
void
MPU680x0Device::mmu_op_pflush_ea()
{
	uint32 ea;
	uint32 mask;
	uint32 fc;

	ea = cea_copro();
	mask = (ir2 >> 5) & 7;
	fc = get_fc();
	atc.flush(fc, mask, ea);
}

// PLOADR/PLOADW 命令
void
MPU680x0Device::mmu_op_pload()
{
	uint32 ea;
	uint32 fc;

	if ((ir2 & 0x01e0) != 0) {
		mmu_op_illg2();
		return;
	}
	ea = cea_copro();
	fc = get_fc();

	// 実効アドレスを論理アドレスとする ATC を作成。
	// ってこれでいいのか?
	if ((ir2 & 0x0200)) {
		fcaddr addr(ea, fcattr::FC(fc) | fcattr::R);
		atc.fill_read(addr);
	} else {
		fcaddr addr(ea, fcattr::FC(fc) | fcattr::W);
		atc.fill_write(addr);
	}
}

// PTESTR/PTESTW 命令
void
MPU680x0Device::mmu_op_ptest()
{
	uint level = (ir2 >> 10) & 7;
	fcattr rw  = (ir2 & 0x0200) ? fcattr::R : fcattr::W;
	uint rn    = (ir2 >> 5) & 0xf;
	uint32 ea;
	uint32 fc;
	uint16 mmusr;

	ea = cea_copro();
	fc = get_fc();
	mmusr = 0;

	if (level == 0) {
		if (rn != 0) {
			// レベル0で An 指定なら Fライン例外
			mmu_op_illg2();
			return;
		}

		// laddr は IS と PS のところを落としたもの。
		// ssw_laddr のアドレスは IS を落とさないもの (PSは不問)。
		fcaddr ssw_laddr(ea, fcattr::FC(fc) | rw);
		uint32 laddr = ea & mmu_tc.lmask;

		const m68030ATCLine *a = atc.fill_peek(laddr, fc);
		if (a) {
			// ATC に見付かった。Modified の状態を返す。
			if (a->IsBusError()) {
				mmusr |= m68030MMUSR::B;
			} else if (a->IsWProtect()) {
				mmusr |= m68030MMUSR::W;
			} else if (a->IsModified()) {
				mmusr |= m68030MMUSR::M;
			}
		} else {
			// ATC に見付からなかった。Invalid を立てる。
			mmusr = m68030MMUSR::I;
		}

		// どちらにしても TT を評価。
		for (int i = 0; i < 2; i++) {
			m68030TT& tt = mmu_tt[i];
			if (tt.Match(ssw_laddr)) {
				mmusr |= m68030MMUSR::T;
			}
		}
	} else {
		// レベル 1-7 ならテーブルサーチを実行。

		bus.laddr.Set(ea, fcattr::FC(fc) | rw);
		m68030MMU table(this);
		table.m_ptest = true;
		table.search(level);

		if (table.m_berr) {
			mmusr |= m68030MMUSR::B;
		}
		if (table.m_lim) {
			// XXX not tested
			mmusr |= m68030MMUSR::L;
		}
		if (table.m_s && (fc & 4)) {
			// XXX not tested
			mmusr |= m68030MMUSR::S;
		}
		if (table.m_wp) {
			mmusr |= m68030MMUSR::W;
		}
		if (table.m_type == INVALID) {
			mmusr |= m68030MMUSR::I;
		}
		if (table.m_m) {
			mmusr |= m68030MMUSR::M;
		}
		// N を求める。
		// もともとそういう風には出来ていないので逆算している。
		int n;
		if (mmu_tc.fcl) {
			// たぶんファンクションテーブルで1段消費する? (未確認)
			if (table.x == FCL) {
				n = 1;
			} else {
				n = table.x + 2;
			}
		} else {
			n = table.x + 1;
		}
		mmusr |= n;

		// アドレスレジスタ指定があれば、最後のディスクリプタアドレスを返す。
		// rn は有効なら 8-15。
		if (rn != 0) {
			// XXX 途中でバスエラーが起きるケースは未実装
			if (table.m_berr == false) {
				reg.R[rn] = table.m_lastaddr;
			}
		}
	}

	SetMMUSR(mmusr);
}
