//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// メインバス (X68030)
//

// IODevice
//   +-- MainbusBaseDevice (InitMainbus() と FC アクセスを持つ)
//   |    +-- MainbusDevice (これがメインバス、システムに1つ)
//   |    |    +-- Mainbus24Device (上位8ビットがテーブルで表せるメインバス)
//   |    |    |    +-- LunaMainbus
//   |    |    |    |    +-- Luna1Mainbus
//   |    |    |    |    +-- Luna88kMainbus
//   |    |    |    +-- NewsMainbus
//   |    |    |
//   |    |    |  +-------------+
//   |    |    +--| X68kMainbus |(上位8ビットが IODevice で表せないため)
//   |    |       +-------------+
//   |    |
//   |    +-- X68kIODevice (FC を受け取るため)
//   |
//   +-- XPbusDevice

#include "mainbus_x68k.h"
#include "bankram.h"
#include "extram.h"
#include "interrupt.h"
#include "mainapp.h"
#include "x68kio.h"

// コンストラクタ
X68kMainbus::X68kMainbus()
	: inherited(OBJ_MAINBUS)
{
	// デバイス
	NEWDV(Interrupt, new X68030Interrupt());
	NEWDV(X68kIO, new X68kIODevice());
	NEWDV(ExtRAM, new ExtRAMDevice());

	// デバイスの割り当ては Init() で行っている。

	// モニタの表示外観は Mainbus24 と同じ。
	monitor.func = ToMonitorCallback(&X68kMainbus::MonitorUpdate);
	monitor.SetSize(75, 33);
	monitor.Regist(ID_MONITOR_MAINBUS);

	monitor_accstat.func =
		ToMonitorCallback(&X68kMainbus::MonitorUpdateAccStat);
	monitor_accstat.SetSize(80, 2 + 16 + 3);
	monitor_accstat.Regist(ID_MONITOR_ACCSTAT);
}

// デストラクタ
X68kMainbus::~X68kMainbus()
{
}

// 初期化
bool
X68kMainbus::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	for (int i = 0; i < bankram.size(); i++) {
		bankram[i] = gMainApp.FindObject<BankRAMDevice>(OBJ_BANKRAM(i));
	}

	return true;
}

// メインバスの初期化
bool
X68kMainbus::InitMainbus()
{
	// 拡張メモリはサイズが確定した後のここで配置。
	extsize = pExtRAM->GetSizeMB();

	// 上位8ビット全域を一旦全部内蔵空間にしてから
	std::fill(is_intio.begin(), is_intio.end(), true);

	if (extsize == 16) {
		// TS-6BE16 互換 (16MB) なら $01 の 16MB。
		extstart = 16;
		is_intio[1] = false;
	} else {
		// 060turbo 互換なら $10 から。
		extstart = 256;
		for (int i = 0x10, end = i + (extsize / 16); i < end; i++) {
			is_intio[i] = false;
		}
	}

	// 親子構造になっている。
	if (pX68kIO->InitMainbus() == false) {
		return false;
	}

	return true;
}

// リセット
void
X68kMainbus::ResetHard(bool poweron)
{
	inherited::ResetHard(poweron);

	std::fill(accstat_read.begin(), accstat_read.end(), 0);
	std::fill(accstat_write.begin(), accstat_write.end(), 0);
}

// リセットについて。
//
//           PWRRESET
// 電源オン ----------+--> VIPS
// (7505)             +--> VideoCtlr
//                    +--> OSCIAN2
//                    |
//                    |   +------+
//                    +-->|      |
// 本体      SWRESET      | YUKI |
// リセット ------------->|      |
// スイッチ               +------+
//                         | IPLRESET
//                         |
//                         +--> FPU
//                         |
//                         ▽
//                   RESET |
//                         |<-> MPU (68030)
//                         +--> PPI (8255)
//                         +--> SPC (MB89352)
//                         +--> ADPCM (MSM6258)
//                         +--> FDC (uPD72065B)
//                         +--> MFP (MC68901)
//                         |
//                         |   +-------+
//                         +-->| PEDEC |
//                         |   +-------+
//                         |     |
//                         |     +-- OPMIC -----> OPM(YM2151)
//                         |     +-- SCCRD/WR --> SCC(Z8530)
//                         |
//                         |      EXRESET
//                         +---|>-----------> 外部バス
//                         ▽
//                SYSRESET |
//                         |<-> SAKI  ---+
//                         |             | BEC0,BEC1
//                         +--> DMAC  <--+
//                         +--> CYNTHIA
// リセットされない?
// o CACHY

// MPU の RESET 命令によるリセット
void
X68kMainbus::ResetByMPU()
{
	std::vector<int> ids {
		OBJ_DMAC,
		OBJ_FDC,
		OBJ_MFP,
		OBJ_PEDEC,
		OBJ_PLUTO,
		OBJ_PPI,
		OBJ_SPC,
	};
	for (auto id : ids) {
		auto dev = gMainApp.GetObject<IODevice>(id);
		dev->ResetHard(false);
	}
}

inline bool
X68kMainbus::IsIntio(uint32 addr) const
{
	return is_intio[(addr >> 24) & 0xff];
}

#define READ(NN)	do {	\
	uint32 addr = (uint32)fcaddr;	\
	if (IsIntio(addr)) {	\
		/* 内蔵 16MB を 1MB 単位で */	\
		accstat_read[(addr >> 20) & 0x0f] = ACC_READ;	\
		return pX68kIO->__CONCAT(Read,NN)(fcaddr);	\
	} else {	\
		/* 拡張メモリのアドレスは折り返しは起きない */	\
		accstat_read[addr >> 20] = ACC_READ;	\
		return pExtRAM->__CONCAT(Read,NN)(addr);	\
	}	\
} while (0)

#define WRITE(NN)	do {	\
	uint32 addr = (uint32)fcaddr;	\
	if (IsIntio(addr)) {	\
		/* 内蔵 16MB を 1MB 単位で */	\
		accstat_write[(addr >> 20) & 0x0f] = ACC_WRITE;	\
		return pX68kIO->__CONCAT(Write,NN)(fcaddr, data);	\
	} else {	\
		/* 拡張メモリのアドレスは折り返しは起きない */	\
		accstat_write[addr >> 20] = ACC_WRITE;	\
		return pExtRAM->__CONCAT(Write,NN)(addr, data);	\
	}	\
} while (0)

uint64
X68kMainbus::Read8(fcaddr fcaddr)
{
	READ(8);
}

uint64
X68kMainbus::Read16(fcaddr fcaddr)
{
	READ(16);
}

uint64
X68kMainbus::Read32(fcaddr fcaddr)
{
	READ(32);
}

uint64
X68kMainbus::Write8(fcaddr fcaddr, uint32 data)
{
	WRITE(8);
}

uint64
X68kMainbus::Write16(fcaddr fcaddr, uint32 data)
{
	WRITE(16);
}

uint64
X68kMainbus::Write32(fcaddr fcaddr, uint32 data)
{
	WRITE(32);
}

uint64
X68kMainbus::Peek8(uint32 addr)
{
	if (IsIntio(addr)) {
		return pX68kIO->Peek8(addr);
	} else {
		return pExtRAM->Peek8(addr);
	}
}

uint64
X68kMainbus::Poke8(uint32 addr, uint32 data)
{
	if (IsIntio(addr)) {
		return pX68kIO->Poke8(addr, data);
	} else {
		return pExtRAM->Poke8(addr, data);
	}
}

void
X68kMainbus::MonitorUpdate(Monitor *, TextScreen& screen)
{
	screen.Clear();

	// 012345678901234567890
	// $0000'0000: 1234567

	for (int i = 0; i < 8; i++) {
		screen.Print(12 + i * 8, 0, "+$0%x", i);
	}

	auto xioname = FormatDevName(pX68kIO.get()).first;
	auto extname = FormatDevName(pExtRAM.get()).first;

	int idx = 0;
	for (int i = 0; i < is_intio.size() / 8; i++) {
		screen.Print(0, i + 1, "$%02x00'0000:", idx);
		for (int j = 0; j < 8; j++) {
			screen.Print(12 + j * 8, i + 1, "%s",
				(is_intio[idx] ? xioname : extname).c_str());
			idx++;
		}
	}
}

void
X68kMainbus::MonitorUpdateAccStat(Monitor *, TextScreen& screen)
{
	std::array<uint8, 1024> acc;
	std::array<uint8, 16 + 16> accbank;

	// R|W をマージしながらコピー。
	for (int i = 0, iend = acc.size(); i < iend; i += 4) {
		uint32 tmp = 0;
		tmp |= *(uint32 *)&accstat_read[i];
		tmp |= *(uint32 *)&accstat_write[i];
		*(uint32 *)&acc[i] = tmp;
	}

	// 読んだのでリセット。
	std::fill(accstat_read.begin(), accstat_read.end(), 0);
	std::fill(accstat_write.begin(), accstat_write.end(), 0);

	// バンクメモリ。
	for (int n = 0; n < bankram.size(); n++) {
		if (bankram[n]) {
			bankram[n]->FetchAccStat(&accbank[n * 16]);
		}
	}

	screen.Clear();
	screen.Print(0, 0,
		"Shows only the first 1GB. 1MB/char. "
		"'R':Read, 'W':Write, 'A':Read+Write");
	screen.Print(12, 1, "0123456789ABCDEF");
	for (int i = 1; i < 4; i++) {
		screen.Print(12 + 17 * i, 1, "+$0%x", i);
	}
	for (int y = 0; y < acc.size() / 64; y++) {
		screen.Print(0, y + 2, "$%02x00'0000:", y * 0x04);
	}

	// 拡張メモリ部分。
	for (int i = extstart, end = extstart + extsize; i < end; i++) {
		uint op = acc[i] & 3;
		uint8 ch = ".RWA"[op];
		int y = 2 + i / 64;
		int x = 12 + (i % 64) + ((i % 64) / 16)/*区切りの空白分*/;
		screen.Putc(x, y, ch);
	}

	// 内蔵I/O 部分。
	{
		std::array<char, 16 + 1> buf;
		char *d = &buf[0];
		for (int i = 0; i < 16; i++) {
			uint op = acc[i] & 3;
			*d++ = ".RWA"[op];
		}
		*d = '\0';
		screen.Puts(12, 2, &buf[0]);
	}

	// バンクメモリ。
	for (int n = 0; n < 2; n++) {
		screen.Print(0, 19 + n, "BankMem #%d:", n);
		if (bankram[n]) {
			for (int i = 0; i < 16; i++) {
				uint op = accbank[n * 16 + i] & 3;
				uint8 ch = ".RWA"[op];
				screen.Putc(12 + i, 19 + n, ch);
			}
		}
	}
}

// ブートページを切り替える。
void
X68kMainbus::SwitchBootPage(bool isrom)
{
	// X68030 ではすべて X68kIO が担当している
	pX68kIO->SwitchBootPage(isrom);
}
